searchNodes=[{"doc":"PoolSup: Yet another process pool library in Elixir PoolSup defines a supervisor specialized to manage pool of worker processes. API Documentation Hex package information Features Process defined by this module behaves as a :simple_one_for_one supervisor. Worker processes are spawned using a callback module that implements PoolSup.Worker behaviour. PoolSup process manages which worker processes are in use and which are not. PoolSup automatically restarts crashed workers. Functions to request pid of an available worker process: checkout/2 , checkout_nonblocking/2 . Run-time reconfiguration of pool size: change_capacity/3 . Automatic cleanup of workers hanging around too long without checkin, as a safeguard against process leaks. Load-balancing using multiple pools: PoolSup.Multi . Example Suppose we have a module that implements both GenServer and PoolSup.Worker behaviours ( PoolSup.Worker behaviour requires only 1 callback to implement, start_link/1 ). defmodule MyWorker do @behaviour PoolSup.Worker use GenServer def start_link ( arg ) do GenServer . start_link ( __MODULE__ , arg ) end # definitions of gen_server callbacks... end When we want to have 3 worker processes that run MyWorker server: { :ok , pool_sup_pid } = PoolSup . start_link ( MyWorker , { :worker , :arg } , 3 , 0 , [ name : :my_pool ] ) Each worker process is started using MyWorker.start_link({:worker, :arg}) . Then we can get a pid of a child currently not in use: worker_pid = PoolSup . checkout ( :my_pool ) do_something ( worker_pid ) PoolSup . checkin ( :my_pool , worker_pid ) Don't forget to return the worker_pid when finished; for simple use cases PoolSup.transaction/3 comes in handy. Reserved and on-demand worker processes PoolSup defines the following two parameters to control capacity of a pool: reserved (3rd argument of start_link/5 ): Number of workers to keep alive. ondemand (4th argument of start_link/5 ): Maximum number of workers that are spawned on-demand. In short: { :ok , pool_sup_pid } = PoolSup . start_link ( MyWorker , { :worker , :arg } , 2 , 1 ) w1 = PoolSup . checkout_nonblocking ( pool_sup_pid ) # Returns a pre-spawned worker pid w2 = PoolSup . checkout_nonblocking ( pool_sup_pid ) # Returns the other pre-spawned worker pid w3 = PoolSup . checkout_nonblocking ( pool_sup_pid ) # Returns a newly-spawned worker pid nil = PoolSup . checkout_nonblocking ( pool_sup_pid ) # Returns `nil`, no available process PoolSup . checkin ( pool_sup_pid , w1 ) # `w1` is terminated PoolSup . checkin ( pool_sup_pid , w2 ) # `w2` is kept alive for the subsequent checkout PoolSup . checkin ( pool_sup_pid , w3 ) # `w3` is kept alive for the subsequent checkout Usage within supervision tree The following code snippet spawns a supervisor that has PoolSup process as one of its children: chilldren = [ ... Supervisor . child_spec ( { PoolSup , [ MyWorker , { :worker , :arg } , 5 , 3 ] } , [ ] ) , ... ] Supervisor . start_link ( children , [ strategy : :one_for_one ] ) The PoolSup process initially has 5 workers and can temporarily have up to 8. All workers are started by MyWorker.start_link({:worker, :arg}) . You can of course define a wrapper function of PoolSup.start_link/4 and use it in your supervisor spec.","ref":"PoolSup.html","title":"PoolSup","type":"module"},{"doc":"Changes capacity (number of worker processes) of a pool. new_reserved and/or new_ondemand parameters can be nil ; in that case the original value is kept unchanged (i.e. PoolSup.change_capacity(pool, 10, nil) replaces only reserved value of pool ). On receipt of change_capacity message, the pool adjusts number of children according to the new configuration as follows: If current number of workers are less than reserved , the pool spawns new workers to ensure reserved workers are available. Note that, as is the same throughout the OTP framework, spawning processes under a supervisor is synchronous operation. Therefore increasing reserved too many at once may make the pool unresponsive for a while. When increasing maximum capacity ( reserved + ondemand ) and if any client process is being checking-out in a blocking manner, then the newly-spawned process is returned to the client. When decreasing capacity, the pool tries to shutdown extra workers that are not in use. Processes currently in use are never interrupted. If number of in-use workers is more than the desired capacity, terminating further is delayed until any worker process is checked in.","ref":"PoolSup.html#change_capacity/3","title":"PoolSup.change_capacity/3","type":"function"},{"doc":"Changes :checkout_max_duration option of the pool. See start_link/5 for detailed explanation of :checkout_max_duration option.","ref":"PoolSup.html#change_checkout_max_duration/2","title":"PoolSup.change_checkout_max_duration/2","type":"function"},{"doc":"Checks in an in-use worker pid and make it available to others.","ref":"PoolSup.html#checkin/2","title":"PoolSup.checkin/2","type":"function"},{"doc":"Checks out a worker pid that is currently not used. If no available worker process exists, the caller is blocked until either any process becomes available, or timeout is reached. Note that when a pid is checked-out it must eventually be checked-in or die, in order to correctly keep track of working processes and avoid process leaks. For this purpose it's advisable to either link the checked-out process and the process who is going to check-in that process, or implement your worker to check-in itself at the end of each job.","ref":"PoolSup.html#checkout/2","title":"PoolSup.checkout/2","type":"function"},{"doc":"Checks out a worker pid in a nonblocking manner, i.e. if no available worker found this returns nil .","ref":"PoolSup.html#checkout_nonblocking/2","title":"PoolSup.checkout_nonblocking/2","type":"function"},{"doc":"Returns a child specification to be used when it's not fully specified by the parent supervisor.","ref":"PoolSup.html#child_spec/1","title":"PoolSup.child_spec/1","type":"function"},{"doc":"Callback implementation for GenServer.format_status/2 .","ref":"PoolSup.html#format_status/2","title":"PoolSup.format_status/2","type":"function"},{"doc":"Callback implementation for GenServer.init/1 .","ref":"PoolSup.html#init/1","title":"PoolSup.init/1","type":"function"},{"doc":"Starts a PoolSup process linked to the calling process. Arguments worker_module is the callback module of PoolSup.Worker . worker_init_arg is the value passed to worker_module.start_link/1 callback function. reserved is the number of workers this PoolSup process holds. ondemand is the maximum number of workers that are spawned on checkouts when all reserved processes are in use. options is a keyword list of the following options: :name : used for name registration for the pool process. :checkout_max_duration : a threshold (in seconds) of worker's checkout duration (see below). Defaults to nil (workers won't be killed). Terminating non-returning workers Sometimes it's difficult to guarantee that a checked-out worker pid will eventually be checked-in. For example there are cases where a caller process of checkout/2 is killed during waiting for a reply (a worker pid) from a pool process, resulting in a process leak of the worker pid. For this purpose PoolSup provides :checkout_max_duration option as a safeguard against process leaks. If a checked-out worker has not been checked-in for longer than :checkout_max_duration seconds, the pool regards the worker process as leaked and kill it. If :checkout_max_duration is nil this cleanup functionality is disabled. You can dynamically change :checkout_max_duration option of a pool by change_checkout_max_duration/2 .","ref":"PoolSup.html#start_link/5","title":"PoolSup.start_link/5","type":"function"},{"doc":"Query current status of a pool.","ref":"PoolSup.html#status/1","title":"PoolSup.status/1","type":"function"},{"doc":"Checks out a worker pid, creates a link to the worker, executes the given function using the pid, and finally checks-in and unlink. The timeout parameter is used only in the checkout step; time elapsed during other steps are not counted.","ref":"PoolSup.html#transaction/3","title":"PoolSup.transaction/3","type":"function"},{"doc":"","ref":"PoolSup.html#t:option/0","title":"PoolSup.option/0","type":"type"},{"doc":"","ref":"PoolSup.html#t:pool/0","title":"PoolSup.pool/0","type":"type"},{"doc":"Defines a supervisor that is specialized to manage multiple PoolSup processes. For high-throughput use cases, centralized process pool such as PoolSup may become a bottleneck as all tasks must checkout from a single pool manager process. This module is for these situations: to manage multiple PoolSup s and load-balance checkout requests to multiple pool manager processes. In summary, Process defined by PoolSup.Multi behaves as a :simple_one_for_one supervisor. Children of PoolSup.Multi are PoolSup processes and they have identical configurations (worker module, capacity, etc.). checkout/3 , checkout_nonblocking/3 and transaction/4 which randomly picks a pool in a PoolSup.Multi (with the help of ETS) are provided. Number of pools and capacity of each pool are dynamically configurable. Example Suppose we have the following worker module: iex(1)&gt; defmodule MyWorker do ...(1)&gt; @behaviour PoolSup.Worker ...(1)&gt; use GenServer ...(1)&gt; def start_link ( arg ) do ...(1)&gt; GenServer . start_link ( __MODULE__ , arg ) ...(1)&gt; end ...(1)&gt; # definitions of gen_server callbacks... ...(1)&gt; end To use PoolSup.Multi it's necessary to setup an ETS table. iex(2)&gt; table_id = :ets . new ( :arbitrary_table_name , [ :set , :public , { :read_concurrency , true } ] ) Note that the PoolSup.Multi process must be able to write to the table. The following creates a PoolSup.Multi process that has 3 PoolSup s each of which manages 5 reserved and 2 ondemand workers. iex(3)&gt; { :ok , pool_multi_pid } = PoolSup.Multi . start_link ( table_id , &quot;arbitrary_key&quot; , 3 , MyWorker , { :worker , :arg } , 5 , 2 ) Now we can checkout a worker pid from the set of pools: iex(4)&gt; { pool_pid , worker_pid } = PoolSup.Multi . checkout ( table_id , &quot;arbitrary_key&quot; ) iex(5)&gt; do_something ( worker_pid ) iex(6)&gt; PoolSup . checkin ( pool_pid , worker_pid )","ref":"PoolSup.Multi.html","title":"PoolSup.Multi","type":"module"},{"doc":"Changes :checkout_max_duration option of child pools. See PoolSup.start_link/5 for detailed explanation of :checkout_max_duration option. The change will be broadcasted to all existing pools. Also all pools that start afterward will use the new value of :checkout_max_duration .","ref":"PoolSup.Multi.html#change_checkout_max_duration/2","title":"PoolSup.Multi.change_checkout_max_duration/2","type":"function"},{"doc":"Changes configuration of an existing PoolSup.Multi process. new_n_pools , new_reserved and/or new_ondemand parameters can be nil ; in that case the original value is kept unchanged. Changing number of pools When new_n_pools is larger than the current number of working pools, PoolSup.Multi spawns new pools immediately. When new_n_pools is smaller than the current number of working pools, PoolSup.Multi process randomly chooses pools to terminate and mark them &quot;not working&quot;, exclude those pools from the ETS record, resets their reserved and ondemand as 0 so that new checkouts will never succeed, starts to periodically poll the status of &quot;not working&quot; pools, and terminate a pool when it becomes ready to terminate (i.e. no worker process is used). Changing reserved and/or ondemand of each pool The given values of reserved , ondemand are notified to all the working pools. See PoolSup.change_capacity/3 for the behaviour of each pool.","ref":"PoolSup.Multi.html#change_configuration/4","title":"PoolSup.Multi.change_configuration/4","type":"function"},{"doc":"Checks out a worker pid that is currently not used. Internally this function looks-up the specified ETS record, randomly chooses one of the pools and checks-out a worker in the pool. Note that this function returns a pair of pid s: {pool_pid, worker_pid} . The returned pool_pid must be used when returning the worker to the pool: PoolSup.checkin(pool_pid, worker_pid) .","ref":"PoolSup.Multi.html#checkout/3","title":"PoolSup.Multi.checkout/3","type":"function"},{"doc":"Checks out a worker pid in a nonblocking manner, i.e. if no available worker found in the randomly chosen pool this returns nil .","ref":"PoolSup.Multi.html#checkout_nonblocking/3","title":"PoolSup.Multi.checkout_nonblocking/3","type":"function"},{"doc":"Returns a child specification to be used when it's not fully specified by the parent supervisor.","ref":"PoolSup.Multi.html#child_spec/1","title":"PoolSup.Multi.child_spec/1","type":"function"},{"doc":"Callback implementation for GenServer.format_status/2 .","ref":"PoolSup.Multi.html#format_status/2","title":"PoolSup.Multi.format_status/2","type":"function"},{"doc":"Callback implementation for GenServer.init/1 .","ref":"PoolSup.Multi.html#init/1","title":"PoolSup.Multi.init/1","type":"function"},{"doc":"Starts a PoolSup.Multi process linked to the calling process. Arguments table_id : ID of the ETS table to use. pool_multi_key : Key to identify the record in the ETS table. Note that PoolSup.Multi keeps track of the PoolSup s within a single ETS record. Thus multiple instances of PoolSup.Multi can share the same ETS table (as long as they use unique keys). n_pools : Number of pools. worker_module : Callback module of PoolSup.Worker . worker_init_arg : Value passed to worker_module.start_link/1 . reserved : Number of reserved workers in each PoolSup . ondemand : Number of ondemand workers in each PoolSup . options : Keyword list of the following options: :name : Used for name registration of PoolSup.Multi process. :checkout_max_duration : An option passed to child pools. See PoolSup.start_link/5 for detail.","ref":"PoolSup.Multi.html#start_link/8","title":"PoolSup.Multi.start_link/8","type":"function"},{"doc":"Picks a pool from the specified ETS record, checks out a worker pid, creates a link to the worker, executes the given function using the pid, and finally checks-in and unlink. The timeout parameter is used only in the checkout step; time elapsed during other steps are not counted.","ref":"PoolSup.Multi.html#transaction/4","title":"PoolSup.Multi.transaction/4","type":"function"},{"doc":"","ref":"PoolSup.Multi.html#t:option/0","title":"PoolSup.Multi.option/0","type":"type"},{"doc":"","ref":"PoolSup.Multi.html#t:pool_multi/0","title":"PoolSup.Multi.pool_multi/0","type":"type"},{"doc":"","ref":"PoolSup.Multi.html#t:pool_multi_key/0","title":"PoolSup.Multi.pool_multi_key/0","type":"type"},{"doc":"","ref":"PoolSup.Multi.html#t:pool_sup_args/0","title":"PoolSup.Multi.pool_sup_args/0","type":"type"},{"doc":"Behaviour definition of worker processes to be managed by PoolSup . To implement PoolSup.Worker behaviour it's enough to implement just one function start_link/1 . The start_link/1 callback function is invoked by PoolSup to spawn a new process. The argument passed to worker's start_link/1 is the second argument of PoolSup.start_link/4 .","ref":"PoolSup.Worker.html","title":"PoolSup.Worker","type":"behaviour"},{"doc":"","ref":"PoolSup.Worker.html#c:start_link/1","title":"PoolSup.Worker.start_link/1","type":"callback"}]